//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2006-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//#define _GNU_SOURCE
#include <stdio.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>

#include "ar_dbus.h"
#include "cmds.h"
#include "common.h"
#include "conf.h"
#include "dispatcher.h"
#include "executor.h"
#include "parse.h"
#include "utils.h"
#include "xemulate.h"

extern char* getBtAddress();		// from btio.h
extern int   getClip 	(char *string); // from atsend.h

extern ConnectInfo connectInfo;
extern char 	    callerId[MAXLEN];
extern type_key*    repeatCmdPtr;
extern char         tmp[MAXMAXLEN];

int macroCmd(char *descr, char *exec, cmdParams* params);

char tmptext[MAXCMDLEN];
int  remoteOn       = 1;

/* Have issues with sockets if use execNoFork()
   If some app started from anyRemote, then it keeps socket open even if anyRemote
   is killed somehow. Have "ERROR: on binding 98->Address already in use" errors because of
   socket CLOSE_WAIT state reported by /usr/sbin/lsof -i :<port> 
   !!! Therefore need to use fork !!!

static int execNoFork(const char *cmd)
{
    DEBUG2("execNoFork >%s<", cmd);

    //printf("execute >%s< command\n",cmd);
    #ifdef __cplusplus
    int rc = 
    #endif
    system(cmd);
    
    //printf("command executed %d\n", rc);

    //if (WIFSIGNALED(rc) && (WTERMSIG(rc) == SIGINT || WTERMSIG(rc) == SIGQUIT)) {
    //     	printf("GOR ABORT in system?");
    //}
    //if(rc!=-1 && WIFEXITED(rc)) {
    //	    rc = WEXITSTATUS(rc);
    //}
    //printf("exit from child %d\n", rc);
    return 0;
}*/

static int execInFork(const char *cmd)
{
    //DEBUG2("execInFork >%s<", cmd);
    int cpid = fork();
    if( cpid < 0) {
	    logger(L_ERR,"Can not fork!");
    	    return -1;
    }

    if(cpid) {      // father
    	    int rf = 0;
    	    waitpid(cpid,&rf,0);
    	    if(rf!=-1 && WIFEXITED(rf)) {
    		    rf = WEXITSTATUS(rf);
    	    }
    	    return rf;
    } 
    
    // else - child
   
    int rc = 0;
    //logger("INF", "Close port in child");
    if (!(connectInfo.mode == CLIENT_RFCOMM || 
          connectInfo.mode == CLIENT_AT)) {
    	    closePort(0);
    }
    
    rc = system(cmd);
    if(rc!=-1 && WIFEXITED(rc)) {
    	    rc = WEXITSTATUS(rc);
    }
    exit(rc);
}

static char* substParams(const char* in, cmdParams *params)
{
    if (in == NULL) {
        logger(L_ERR,"substParams(): null input");
        return NULL;
    }
    DEBUG2("substParams >%s<", in);

    if (*in == '\0') { // Empty command. Skip it.
        return NULL;
    }

    char *command =  strdup(in);
    char *ptr     = command;
    char *papameterized = NULL;

    int pc = strlen(PARAM_CALLID);
    int pm = strlen(PARAM_MODE);
    int pp = strlen(PARAM_PARAM);
    int pi = strlen(PARAM_INDEX);
    //int pd = strlen(PARAM_CFGDIR);
    //int ph = strlen(PARAM_HOME);
    int pb = strlen(PARAM_BTADDR);

    // calculate size to allocate
    int needSize = strlen(ptr)+1;
    int nParams  = 0;
    while(ptr != NULL) {
        papameterized = strstr(ptr, PARAM_START);   // $(

        if (papameterized == NULL) { // Just copy
            break;
        } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
            needSize += strlen(callerId) - pc - 2;
            ptr = papameterized + pc + 2;
            nParams++;
        } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
            char *m = getMode();
            needSize += strlen(m) - pm - 2;
            ptr = papameterized + pm + 2;
            nParams++;
        } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
            if (params != NULL && params->value != NULL) {   // value can be empty
                needSize += strlen(params->value) - pp - 2;
            }
            ptr = papameterized + pp + 2;
            nParams++;
        } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
            if (params != NULL && params->index != NULL && params->index[0] != '\0') {
                needSize += strlen(params->index) - pi - 2;
            }
            ptr = papameterized + pi + 2;
            nParams++;
            /*} else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {
            int c = 0;
                                 char *d;
                                 if ((d = getCfgDir()) != NULL) {
            			c = strlen(d);
                                         needSize += c - pd - 2;
                                 }
                                 ptr = papameterized + pd + 2;
                                 nParams++;
                         } else if (strncmp(papameterized + 2,PARAM_HOME,ph) == 0) {
            		int c = 0;
                                 char *d;
                                 if ((d = getenv("HOME")) != NULL) {
            			c = strlen(d);
                                         needSize += c - ph - 2;
                                 }
                                 ptr = papameterized + ph + 2;
                                 nParams++;*/
        } else if (strncmp(papameterized + 2,PARAM_BTADDR,pb) == 0) {
            int c = 0;
            char *d;
            if ((d = getBtAddress()) != NULL) {
                c = strlen(d);
                needSize += c - pb - 2;
            }
            ptr = papameterized + pb + 2;
            nParams++;
        } else {
            int pLen = (papameterized == NULL ? 0 : 1);
            if (pLen) {
                char * br = index(papameterized+2,')');
                pLen = (br == NULL ? 0 : br - papameterized - 2);
            }

            varData* var = NULL;

            if ((var = searchVar(papameterized + 2, pLen)) != NULL) {

                //logger(L_DBG, "searchVar() OK");

                needSize += (var == NULL ? 0 : var->size - strlen(var->name) - 3);

                //sprintf(tmp, "Add bytes for var: %d", (int) (var->size - strlen(var->name) - 3));
                //logger(L_DBG,tmp);

                ptr = papameterized + strlen(var->name) + 3;
                nParams++;
            } else {
                break;
            }
        }
    }

    if (nParams == 0) {		// non-parametrized command
        return command;
    }

    papameterized = NULL;
    ptr = command;

    //DEBUG2("Allocate >%d< bytes for parametrized command", needSize);

    char* cmd  = (char*) calloc(needSize+1,1);
    int  count = 0;

    while(ptr != NULL) {
        DEBUG2("current cmd >%s<", ptr);

        papameterized = strstr(ptr, PARAM_START);   // $(

        if (papameterized != NULL) {
            DEBUG2("papameterized >%s<", papameterized);
        }

        if (papameterized == NULL) { // Just copy
            logger(L_DBG, "no more parameters");
            strcpy(cmd+count, ptr);
            DEBUG2("cur cmd >%s<", cmd);

            break;
        } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {

            *papameterized = '\0';

            strcpy(cmd+count, ptr);		// First part
            strcat(cmd+count, callerId);    // Add params

            count += strlen(ptr) + strlen(callerId);
            ptr = papameterized + pc + 2;

        } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {

            *papameterized = '\0';
            char *m = getMode();

            strcpy(cmd+count, ptr); 	// First part
            strcat(cmd+count, m);		// Add params

            count += strlen(ptr) + strlen(m);
            ptr = papameterized + pm + 2;


        } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
            // This is a trick to make Bemused commands like VOLM<volume> works
            // also it is used in Java Client list handling

            *papameterized = '\0';

            if (params == NULL) {   // value can be empty
                logger(L_WARN, "No parameters supplied!");

                strcpy(cmd+count, PARAM_START);
                strcat(cmd+count, PARAM_PARAM);

                count += strlen(PARAM_START) + strlen(PARAM_PARAM);
            } else {
                strcpy(cmd+count, ptr);			// First part
                strcat(cmd+count, params->value);	// Add params

                count += strlen(ptr) + strlen(params->value);
            }
            ptr = papameterized + pp + 2;

        } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
            // Used in Java Client list handling

            *papameterized = '\0';

            if (params == NULL || params->index == NULL || params->index[0] == '\0') {
                logger(L_WARN, "No parameters supplied!");

                strcpy(cmd+count, PARAM_START);
                strcat(cmd+count, PARAM_INDEX);

                count += strlen(PARAM_START) + strlen(PARAM_INDEX);
            } else {
                strcpy(cmd+count, ptr);			// First part
                strcat(cmd+count, params->index);	// Add params

                count += strlen(ptr) + strlen(params->index);
            }
            ptr = papameterized + pi + 2;

            /*} else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {

                    *papameterized = '\0';

                    // First part
                    strcpy(cmd+count, ptr);
            	count += strlen(ptr);

                    // Add cfg. dir.
                    char *d;
                    if ((d = getCfgDir()) != NULL) {
            strcpy(cmd+count, d);
                    	count += strlen(d);
                    }

                    ptr = papameterized + pd + 2;

            } else if (strncmp(papameterized + 2,PARAM_HOME,ph) == 0) {

                    *papameterized = '\0';

                    // First part
                    strcpy(cmd+count, ptr);
            	count += strlen(ptr);

                    // Add $HOME
                    char *d;
                    if ((d = getenv("HOME")) != NULL) {
            strcpy(cmd+count, d);
                    	count += strlen(d);
                    }

                    ptr = papameterized + ph + 2;
            */
        } else if (strncmp(papameterized + 2,PARAM_BTADDR,pb) == 0) {

            *papameterized = '\0';

            // First part
            strcpy(cmd+count, ptr);
            count += strlen(ptr);

            // Add XX:XX:XX:XX:XX:XX
            char *d;
            if ((d = getBtAddress()) != NULL) {
                strcpy(cmd+count, d);
                count += strlen(d);
            }

            ptr = papameterized + pb + 2;

        } else {

            int pLen = (papameterized == NULL ? 0 : 1);
            if (pLen) {
                char * br = index(papameterized+2,')');
                pLen = (br == NULL ? 0 : br - papameterized - 2);
            }

            varData* var = NULL;

            if ((var = searchVar(papameterized + 2, pLen)) != NULL) {
                logger(L_DBG, "substitute: searchVar() OK");

                *papameterized = '\0';
                strcpy(cmd+count, ptr);		// First part
                count += strlen(ptr);

                if (var->size > 0) {
                    memcpy(cmd+count, var->value, var->size);	// Add params
                    count += var->size;
                }

                DEBUG2("cmd var value >%s<", cmd);

                ptr = papameterized + strlen(var->name) + 3;

            } else {    // Just copy
                //sprintf(tmp, "Just copy >%s<", ptr);
                //logger(L_DBG,tmp);
                strcpy(cmd+count, ptr);
                count += strlen(ptr);
                break;
            }
        }
    }
    DEBUG2("final cmd >%s< %d", cmd,count);

    free(command);

    return cmd;
}

#define RSIZE 4096

static GByteArray* executeCommandPipe(const char* exec)
{
    //DEBUG2("executeCommandPipe %s",exec);
   
    FILE *fp = popen(exec, "r");
    if (fp == NULL) {
        logger(L_ERR, "Error on popen()");
	return NULL;
    }

    GByteArray* data = g_byte_array_new();
    
    guint8 temp[RSIZE];
    while (1) {
        size_t num = fread(temp, 1, RSIZE, fp);
	g_byte_array_append(data, temp, num);
	
	if (num < RSIZE) break;
    }
    

    /*int status = */
    
    pclose(fp);
    
    /*if (status == -1) {
	// Error reported by pclose() 
	//...
    } else {
	// Use macros described under wait() to inspect `status' in order
	//   to determine success/failure of command executed by popen()
	//...
    }*/
     
    //DEBUG2("executeCommandPipe GOT >%s<",data->data);
    return data;
}

static int executeSimpleCommand(const char* exec, char** res, int* resLen)
{
    if (exec == NULL || strlen(exec) == 0 || remoteOn != 1) {
        logger(L_DBG, "executeSimpleCommand null input or remote if OFF");
        return EXIT_OK;
    }
    
    //DEBUG2("executeSimpleCommand >%s<", exec);

    GByteArray* gstr = executeCommandPipe(exec);
    if ( !gstr) return EXIT_NOK;
    
    (*res) = (char*) calloc(gstr->len,1);
    memcpy((*res), gstr->data, gstr->len);
    (*resLen) = gstr->len;
    //DEBUG2("executeSimpleCommand RET >%s<", *res);
   
    g_byte_array_free(gstr, TRUE);
    return EXIT_OK;
}

static char* executeCommand(const char* exec, int* resLen, cmdParams *params)
{
    //logger(L_DBG,"executeCommand");

    char* sExec = substParams(exec, params);
    if (sExec == NULL) {
        logger(L_DBG, "Empty exec. part after substParams()");
        return NULL;
    }
    
    char *res = NULL;
    executeSimpleCommand(sExec, &res, resLen);
    free(sExec);

    return res;
}

static int executeSimpleCmdNoPipe(const char *cmd)
{
    if (cmd == NULL || strlen(cmd) == 0 || remoteOn != 1) {
        logger(L_DBG, "executeSimpleCmdNoPipe null input or remote if OFF");
        return EXIT_OK;
    }
    DEBUG2("executeSimpleCmdNoPipe >%s<", cmd);

    char* p = (char*) calloc(strlen(cmd) + 2,1);
    if (p == NULL) {
        return EXIT_NOK;
    }
    strcpy(p,cmd);

    // How to handle & inside file names ?
    //if (!index(p, '&') ) {

    int idx = strlen(cmd)-1;
    while (idx > 0 && isspace(cmd[idx])) {
    	idx--;
    }

    if (cmd[idx] != '&') {
    	strcat(p, "&");
    }

    //int ret = execNoFork(p);
    int ret = execInFork(p);

    free(p);
    return (ret == -1 ? EXIT_NOK : EXIT_OK);
}


static int processSendCmd(char* sDescr)
{
    int ret = EXIT_OK;

    char all[MAXMAXLEN];
    memset(all,0,MAXMAXLEN);

    strncpy(all, sDescr, MAXMAXLEN-1);

    char *tag  = strtok(all, ",");
    if (tag == NULL) {
        logger(L_DBG, "processSendCmd(): incorrect input (1)");
        return EXIT_NOK;
    }
    if (strlen(sDescr) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        logger(L_DBG, "processSendCmd(): incorrect input (2)");
        return EXIT_NOK;
    }
    char *ptr   = tag + strlen(tag) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
        *ptr = '\0';
        ptr--;
    }
    ptr = tag + strlen(tag) + 1;
    // strip spaces from head of value
    while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        ptr++;
    }

    DEBUG2("processSendCmd() is >%s< >%s<",tag,ptr);

    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->size  = 1;
    dm->file  = NULL;
    if (strcmp(tag, SEND_STRING) == 0) {

        dm->type  = DM_SENDS;
        dm->size  = strlen(ptr);

    } else if (strcmp(tag, SEND_BYTES) == 0) {

        dm->type  = DM_SENDB;

    } else {
        logger(L_DBG, "processSendCmd(): incorrect input (3)");
        free(dm);
        return EXIT_NOK;
    }

    dm->value = (void*) strdup(ptr);
    sendToDispatcher(dm);

    return ret;
}

static void sendData(int type, int flag, char *data, int count)
{
    DEBUG2("sendData() >%d<",count);

    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->size  = count;
    dm->file  = NULL;

    if (type == ID_SEND && flag) {
        dm->type  = DM_SENDB;
    } else if (count > 0) { // just check count to be sure we goes here from right place
        dm->type  = DM_SET;
    }

    dm->value = (void*) strdup(data);
    sendToDispatcher(dm);
}

int execDynamically(char *command)
{
    logger(L_DBG,"execDynamically");
    cmdItem* tmpCmd = NULL;
    storeCmds(&tmpCmd, command); // parse command on-the-fly

    int ret = processCommands(tmpCmd,NULL);
    freeCmds(tmpCmd);

    return ret;
}

static int loadCmds(char *file)
{
    if (file == NULL) {
        return EXIT_NOK;
    }
    DEBUG2("loadCmds %s", file);

    FILE *fp=fopen(file,"r");
    if (fp == NULL) {
        return EXIT_NOK;
    }

    struct stat file_status;
    stat(file, &file_status);

    if (file_status.st_size <= 0) {
        logger(L_DBG,"loadCmds - file is empty");
        fclose(fp);
        return EXIT_NOK;
    }

    char *command = (char*) malloc(file_status.st_size+1);
    *command = '\0';

    int ret = EXIT_OK;
    while (fgets(command, file_status.st_size+1, fp) != NULL) {
        DEBUG2("Retrieved line %s of length %d ", command, (int) strlen(command));
        ret = execDynamically(command);
    }
    free(command);
    fclose(fp);
    return ret;
}

int handleHook(int hook)
{
    DEBUG2("handleHook() >%d<",hook);
    char hookFile[MAXLEN];
    strcpy(hookFile, getenv("HOME"));
    strcat(hookFile, "/.anyRemote/");

    if (hook == ID_EVT_INIT) {
        strcat(hookFile, "hook_init");
    } else if (hook == ID_EVT_CONNECT) {
        strcat(hookFile, "hook_connect");
    } else if (hook == ID_EVT_DISCONNECT) {
        strcat(hookFile, "hook_disconnect");
    } else if (hook == ID_EVT_EXIT) {
        strcat(hookFile, "hook_exit");
    } else {
        return EXIT_NOK;
    }
    return loadCmds(hookFile);
}

static int processTimerCmd(cmdItem* ci)
{
    if (ci->exec == NULL && ci->tdata == NULL) {
        logger(L_DBG, "processTimerCmd(): nothing to exec or timer data absent");
        return EXIT_NOK;
    }

    if (ci->tdata == NULL) {	// Timer(key,cancel|pause|continue|reset|restart)
        timerCmd *tm = findTimer(ci->descr, NULL);
        if ( tm == NULL) {
            logger(L_DBG, "processTimerCmd(): timer did not exists");
            return EXIT_NOK;
        } else {
            sprintf(tmp, "processTimerCmd() change status of %s timer",ci->descr);
            logger(L_DBG, tmp);

            if (strcmp(ci->exec,"pause") == 0) {
                logger(L_DBG, "processTimerCmd(): new status is paused");
                tm->status = TIMER_PAUSE;
            } else if (strcmp(ci->exec,"cancel") == 0) {
                logger(L_DBG, "processTimerCmd(): cancel this timer");
                cancelTimer(ci->descr);
            } else if (strcmp(ci->exec,"continue") == 0) {
                logger(L_DBG, "processTimerCmd(): new status is running");
                tm->status = TIMER_RUN;
            } else if (strcmp(ci->exec,"reset") == 0) {
                logger(L_DBG, "processTimerCmd(): reset timer");
                tm->ticks  = 0;
            } else if (strcmp(ci->exec,"restart") == 0) {
                logger(L_DBG, "processTimerCmd(): restart timer");
                tm->ticks  = 0;
                tm->times  = 1;
            } else {
                logger(L_DBG, "processTimerCmd(): incorrect status");
            }
        }
    } else {
        timerCmd* tc = NULL;
        if ((tc = createTimer(ci)) == NULL) {
            logger(L_DBG, "processTimerCmd(): timer already exists or input was incorrect");
            return EXIT_NOK;
        }

        /* execute it only after timeout
        macroCmd(ci->descr,NULL,NULL);
        	*/
    }
    return EXIT_OK;
}

extern mode       *currentMode;

static int includeCmd(char *file)
{
    if (file == NULL) {
        return EXIT_NOK;
    }

    int ret = EXIT_OK;

    if (load_cfg(file, 0) == -1) {
        logger(L_ERR,"Can not include file !");
    }

    printConf();
    
    return ret;
}

static int setModeCmd(const char *sDescr)
{
    int i   = 0;
    int ret = EXIT_OK;

    type_key* k = findItem(EVT_EXITMODE, &i, NULL);
    if (k && i == FLAG_EXACT) {
        logger(L_DBG, "Exec cmd on exit from mode");
        ret = handleCmdByKey(k,NULL);
        if (ret == EXIT_STOP) {
            return ret;
        }
    } else {
        logger(L_DBG, "Leave mode");
    }
    setMode(sDescr);

    i = 0;
    k = findItem(EVT_ENTRMODE, &i, NULL);
    if (k && i == FLAG_EXACT) {
        logger(L_DBG, "Exec cmd on enter to mode");
        ret = handleCmdByKey(k,NULL);
    } else {
        logger(L_DBG, "Entered to mode");
    }
    return ret;
}

static int setAlarmCmd(const char *sDescr)
{
    char*dsc = strdup(sDescr);

    char* file  = strtok(dsc,",");
    char* macro = strtok(NULL,",");

    if (file != NULL && macro != NULL) {
        addAlarm(file, macro);
        free(dsc);
        return EXIT_OK;
    }
    free(dsc);
    return EXIT_NOK;
}

static int getCmd(const char *sDescr)
{
    if (sDescr == NULL) {
        return EXIT_NOK;
    }

    strcpy(tmptext, CMD_GET);
    strcat(tmptext, "(");
    strcat(tmptext, sDescr);
    strcat(tmptext, ");");

    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->size  = strlen(tmptext);
    dm->value = strdup(tmptext);
    dm->file  = NULL;

    if (connectInfo.mode == SERVER_BT  ||
        connectInfo.mode == SERVER_TCP ||
        connectInfo.mode == SERVER_WEB ||
        connectInfo.mode == SERVER_UX  ||
        connectInfo.mode == CLIENT_NOAT) {

        dm->type  = DM_SET;

#ifdef USE_L2CAP
    } else if (connectInfo.mode == SERVER_L2CAP) {

        dm->type  = DM_SETL2;
#endif
    } else {
        free(dm);
        free(dm->value);
        return  EXIT_NOK;
    }

    sendToDispatcher(dm);
    return  EXIT_OK;
}

static int makeCmd(cmdItem* ci, cmdParams* p)
{
    int ret = EXIT_OK;

    if (ci->descr == NULL) {
        logger(L_DBG, "Empty command description");
        return EXIT_NOK;
    }

    if (strstr(ci->descr,MAKE_VAR) == ci->descr) {

        logger(L_INF, "Command: Make(var,...)");

        char *eValue = ci->exec;
        if (ci->exec && strstr(ci->exec,"by_value") == ci->exec) {
            eValue = (ci->exec + 9);
        }

        char* sExec = substParams(eValue, p);
        //printf("AAA: %s\n", sExec ? sExec : "NULL");

        int resLen = -1;
        char *res  = NULL;

        if (eValue == ci->exec) {
            if(sExec == NULL) {
                logger(L_DBG, "Empty Make(var,...) command after substParams()");
                return EXIT_NOK;
            }

 	    int rcode = executeSimpleCommand(sExec, &res, &resLen);
	    if (rcode == EXIT_NOK) {
	        free(sExec);
		return EXIT_NOK;
            }
            if (*(res+resLen-1) == '\n') {
                resLen--;
            }
        } else {	// set "by_value"
            if (sExec) {
                resLen = strlen(sExec);
                res    = strdup(sExec);
            } else {
                resLen = 0;
            }
        }

        ret = setVar(ci->descr+4,res,resLen);
        if (ret != EXIT_OK) {
            logger(L_DBG,"Fails in setVar() (Make)");
        }

        res = NULL;
        free(sExec);
        sExec = NULL;

    } else if (strstr(ci->descr,MAKE_MODE) == ci->descr) {

        logger(L_INF, "Command: Make(mode,...)");

        char* sExec = substParams(ci->exec, p);

        ret = setModeCmd(sExec);

        free(sExec);
        sExec = NULL;

    } else if (strstr(ci->descr,MAKE_ALARM) == ci->descr) {

        logger(L_INF, "Command: Make(alarm,...)");

        char* sExec = substParams(ci->exec, p);

        ret = setAlarmCmd(sExec);

        free(sExec);
        sExec = NULL;

    } else if (strstr(ci->descr,MAKE_REMOTE) == ci->descr) {

        if (ci->exec && strcmp(ci->exec,"on")==0) {
            logger(L_INF, "Command: Make(remote,on)");
            remoteOn = 1;
        } else if (ci->exec && strcmp(ci->exec,"off")==0) {
            logger(L_INF, "Command: Make(remote,off)");
            remoteOn = 0;
        }

    } else if (strstr(ci->descr,MAKE_DISCONN) == ci->descr) {

        logger(L_INF, "Temporary disconnect command processed");

	freeCachedData();
	
        dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
        dm->size  = 0;
        dm->value = (void*) NULL;
        dm->type  = DM_EVENT_DISCONNECT;
	dm->file  = NULL;
	
        sendToDispatcher(dm);

        return EXIT_OK; // But what with the rest of the command line ? (now just skip)

    } else if (strstr(ci->descr,MAKE_FLUSH) == ci->descr) {

        logger(L_INF, "Command: Make(flush)");
        ret = flushData();

    } else if (strstr(ci->descr,MAKE_STOP) == ci->descr) {

        logger(L_INF, "Command: Make(stop)");
        return EXIT_STOP;

    } else if (strstr(ci->descr,MAKE_EXIT) == ci->descr) {

        logger(L_INF, "Exit command processed");
        sendAbort();
        return EXIT_OK;

        //} else if (strstr(ci->descr,MAKE_NONE) == ci->descr) {
        //        ;
    }

    return ret;
}

static int ckpdCmd(const char *sDescr)
{
    if (connectInfo.mode == CLIENT_RFCOMM || connectInfo.mode == CLIENT_AT) {

        dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
        dm->type  = DM_CKPD;
        dm->size  = strlen(sDescr);
        dm->value = (void*) strdup(sDescr);
	dm->file  = NULL;

        sendToDispatcher(dm);
    }
    return  EXIT_OK;
}

/*static void uploadPixOld(char *cmdLine)
{
    DEBUG2("uploadPix >%s<", cmdLine);
    int prefixSz = 0;          		// image,icon,_name_,_file_|image,window,_file_|cover, _file_

    if (strstr(cmdLine,UPLOAD_PIX)   != NULL) {
        prefixSz = strlen(UPLOAD_PIX) + 1;
    } else if (strstr(cmdLine,UPLOAD_COVER) != NULL) {
        prefixSz = strlen(UPLOAD_COVER) + 1;
    } else {
        int nCh = strlen(UPLOAD_ICON) + 1;
        prefixSz = nCh + strstr(cmdLine+nCh,",") - (cmdLine+nCh) + 1;
    }

    char* fname = cmdLine + prefixSz;
    char* p = fname;
    while (*p != '\0') {	// remove trailing \n if exists
        if (*p == '\n') {
            *p = '\0';
            break;
        }
        p++;
    }
    //sprintf(tmp, "uploadPix file >%s<", fname);
    //logger(L_INF,tmp);

    if (strstr(cmdLine,UPLOAD_COVER) != NULL &&
            isDataNew(ID_COVER, fname, strlen(fname)) == EXIT_NOK) {

        logger(L_INF,"Skip upload same cover");
        return;
    }

    FILE *fp;
    struct stat buf;

    long fLen = 0;
    if(stat(fname, &buf) == -1) {
        logger(L_ERR,"can't get file size!");
        if (strstr(cmdLine,UPLOAD_COVER) == NULL) {	// in case UPLOAD_COVER we will send empty image to clear previous one
            return;
        } else {
            logger(L_DBG,"uploadPix(): send empty cover");

            dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
            dm->type  = DM_SET;
            dm->size  = 11;
            dm->value = (void*) strdup("Set(cover);");

            sendToDispatcher(dm);

            return;
        }
    } else {
        fLen = buf.st_size;
    }

    if (!S_ISREG (buf.st_mode)) {
        logger(L_ERR,"not regular file");
        return;
    }

    //sprintf(tmp, "uploadPix >%ld<", fLen);
    //logger(L_DBG,tmp);

    fp=fopen(fname,"r");
    if (fp == NULL) {
        logger(L_ERR,"can't open file!");
        return;
    }

    uint32_t szh32 = (uint32_t) fLen;
    uint32_t szi32 = htonl(szh32);

    char * fBuffer = (char*) calloc(fLen+prefixSz+12,1);	// 11 =  "Set("     ,<size of bin data=4bytes>    ");\0"
    if (fBuffer == NULL) {
        logger(L_ERR,"no more memory!");
        return;
    }

    strcpy(fBuffer, "Set(");
    strncat(fBuffer, cmdLine, prefixSz+1);
    memcpy((void*)fBuffer+prefixSz+4, (const void *) &szi32, 4);	// length on binary data

    if (fp) {
        #ifdef __cplusplus
        size_t dummy =  
        #endif
        fread(fBuffer+prefixSz+8, sizeof(char), fLen, fp);
	
        fclose(fp);
    }
    strcpy(fBuffer+prefixSz+fLen+8,");");

    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->type  = DM_SET;
    dm->size  = prefixSz+fLen+10;
    dm->value = (void*) fBuffer;	// transfer ovnership to dispatcher

    sendToDispatcher(dm);

    return;
}*/

static void uploadPix(char *cmdLine)
{
    DEBUG2("uploadPix >%s<", cmdLine);
    int prefixSz = 0;          		// image,icon,_name_,_file_|image,window,_file_|cover, _file_

    if (strstr(cmdLine,UPLOAD_PIX)   != NULL) {
        prefixSz = strlen(UPLOAD_PIX) + 1;
    } else if (strstr(cmdLine,UPLOAD_COVER) != NULL) {
        prefixSz = strlen(UPLOAD_COVER) + 1;
    } else {
        int nCh = strlen(UPLOAD_ICON) + 1;
        prefixSz = nCh + strstr(cmdLine+nCh,",") - (cmdLine+nCh) + 1;
    }

    char* fname = cmdLine + prefixSz;
    char* p = fname;
    while (*p != '\0') {	// remove trailing \n if exists
        if (*p == '\n') {
            *p = '\0';
            break;
        }
        p++;
    }
    sprintf(tmp, "uploadPix file >%s<", fname);
    
    logger(L_INF,tmp);

    if (strstr(cmdLine,UPLOAD_COVER) != NULL &&
        isDataNew(ID_COVER, fname, strlen(fname)) == EXIT_NOK) {

        logger(L_INF,"Skip upload same cover");
        return;
    }

    char * fBuffer = (char*) calloc(prefixSz+6,1); // "Set(" + "," + "\0"
    if (fBuffer == NULL) {
        logger(L_ERR,"no more memory!");
        return;
    }
    strcpy(fBuffer, "Set(");
    strncat(fBuffer, cmdLine, prefixSz);
    
    char * fName = (char*) calloc(strlen(fname)+1,1);
    if (fName == NULL) {
        logger(L_ERR,"no more memory!");
        return;
    }
    strcpy(fName,fname);

    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->type  = DM_SETFILE;
    dm->size  = -1;
    dm->value = (void*) fBuffer;	// transfer ovnership to dispatcher
    dm->file  = (void*) fName;	

    sendToDispatcher(dm);

    return;
}

int macroCmd(char *descr, char *exec, cmdParams* params)
{
    if (descr == NULL) {
        return EXIT_NOK;
    }

    char* sDescr = substParams(descr, params);
    if (sDescr == NULL) {
        logger(L_DBG, "Empty macro name in Macro(name,...) command after substParams()");
        return EXIT_NOK;
    }
    DEBUG2("macroCmd/timerCmd %s", sDescr);

    int runIt = 1;
    int ret = EXIT_OK;

    if (exec != NULL) {
        int resLen = -1;

        /*if (exec && strstr(exec,"if_mode_at") == exec) {
        if (connectInfo.mode != CLIENT_AT && connectInfo.mode != CLIENT_RFCOMM) {
                             	runIt = 0;
                             }
                     } else if (exec && strstr(exec,"if_mode_server") == exec) {
        		if (connectInfo.mode != SERVER_BT  &&
        		    #ifdef USE_L2CAP
                                 connectInfo.mode != SERVER_L2CAP &&
        		    #endif
                                 connectInfo.mode != SERVER_TCP &&
                                 connectInfo.mode != SERVER_UX  &&
                                 connectInfo.mode != CLIENT_NOAT) {
                             	runIt = 0;
                             }
                     } else if (exec && strstr(exec,"if_mode_stdin") == exec) {
        		if (connectInfo.mode != SERVER_STDIN) {
                             	runIt = 0;
                             }
                     } else if (exec && strstr(exec,"if_mode_ilirc") == exec) {
        		if (connectInfo.mode != CLIENT_ILIRC) {
                             	runIt = 0;
                             }
        	} else {*/


        char* eValue = exec;
        if (exec && strstr(exec,"by_value") == exec) {
            eValue = (exec + 9);
        }

        char* sExec = substParams(eValue, params);
        char* res  = NULL;

        if (eValue == exec) {
            if(sExec == NULL) {
                logger(L_DBG, "Empty command in Macro(name,...) command after substParams()");
                runIt = 0;
            } else {
		
		int rcode = executeSimpleCommand(sExec, &res, &resLen);
		if (rcode == EXIT_NOK) {
		    free(sExec);
		    free(sDescr);
		    return EXIT_NOK;
                }
                if (*(res+resLen-1) == '\n') {
                    resLen--;
                }
            }
        } else {	// set "by_value"
            if (sExec) {
                resLen = strlen(sExec);
                res    = strdup(sExec);
            } else {
                resLen = 0;
            }
        }


        if (res == NULL) {
            runIt = 0;
        } else {
            if (strcmp(res,"0") != 0 && strcmp(res,"0\n") != 0) {	// test return code $?
                runIt = 0;
            }
            res = NULL;
        }
        free(sExec);
        sExec = NULL;
        //}
    }

    if (runIt == 0) {
        logger(L_DBG,"macroCmd condition not met");
	free(sDescr);
        return EXIT_OK;
    }

    int found;
    type_key *k = findItem(sDescr, &found, NULL);

    if (k && found == FLAG_EXACT) {
        ret = handleCmdByKey(k, params);
        DEBUG2("Return code from macroCmd -> handleCmdByKey() is %d",ret);
    }
    free(sDescr);

    return ret;
}

static int execCmdAndGetResults(int type, cmdItem* ci, timerCmd* timer, cmdParams *params)
{
    //logger(L_DBG,"execCmdAndGetResults");

    if (connectInfo.mode == CLIENT_RFCOMM ||
        connectInfo.mode == CLIENT_AT     ||
        connectInfo.mode == CLIENT_ILIRC  ||
            remoteOn != 1 || ci->exec == NULL) {
        return EXIT_OK;
    }

    //printf("TRACE %s->%s\n",ci->descr,(ci->exec ? ci->exec : "NULL"));

    char* sDescr = substParams(ci->descr, params);
    if (sDescr == NULL) {
        logger(L_DBG, "Empty command after substParams()");
        return EXIT_NOK;
    }

    int resLen = -1;
    char *res = executeCommand(ci->exec, &resLen, params);
    if (res == NULL) {
        logger(L_ERR, "executeCommand() returns NULL");
        free(sDescr);
        return EXIT_NOK;
    }

    if (type == ID_SEND) {  // Just send string

        int bytesFlag = 0;
        if (strstr(sDescr, SEND_BYTES)) {
            bytesFlag = 1;
        }
        sendData(type, bytesFlag, res, resLen);

    } else {                // ID_SET
        //logger(L_DBG, "ID_SET");

        if (strstr(sDescr,UPLOAD_ICON)   == sDescr ||
            strstr(sDescr,UPLOAD_PIX)	 == sDescr ||
            strstr(sDescr,UPLOAD_COVER)  == sDescr) {

            char *chunk = (char*) calloc(strlen(sDescr)+resLen+2,1);

            strcpy(chunk,sDescr);
            strcat(chunk,",");
            memcpy(chunk+strlen(sDescr)+1,res,resLen);

            uploadPix(chunk);

            free (chunk);

        } else {

            if (isDataNew(cmdSet2id(sDescr), res, resLen)) {
                logger(L_INF, "Skip to send the same data");
            } else {
                //logger(L_INF, "Data is new");

                // Replace all ");" inside buffer to "@@" since Java Client will separate commands by ");"
                char *p;
                while ((p = strstr(res,");")) != NULL) {	// binary data ???
                    *p = '@';
                    p++;
                    *p = '@';
                }

                int memSize = resLen + strlen(sDescr) + 7;
                char *chunk = (char*) calloc(1,memSize+1);
                strcpy(chunk,CMD_SET);

                strcat(chunk,"(");  // For example, Set(title
                strcat(chunk,sDescr);

                strcat(chunk,",");
                memcpy(chunk+5+strlen(sDescr),res,resLen);	// could be binary data here
                strcpy(chunk+5+strlen(sDescr)+resLen,");");

                //printf("CHUNK: %s\n", chunk);

                sendData(type, 0, chunk, memSize);

                free (chunk);
            }
        }
    }

    free(sDescr);

    //logger(L_INF,"execCmdAndGetResults finished");
    return EXIT_OK;
}

static int processOneCommand(cmdItem* ci, cmdParams* p)
{
    int ret = EXIT_OK;

    INFO2("processOneCommand >%s< >%s< >%d<",
          ci->descr == NULL ? "no description" : ci->descr,
          ci->exec == NULL  ? "no exec" :ci->exec, remoteOn);

    if (ci->type == ID_MAKE &&
            strstr(ci->descr,MAKE_REMOTE) == ci->descr &&
            strcmp(ci->exec,"on") == 0) {

        logger(L_INF, "Switch remote on");
        remoteOn = 1;
        return EXIT_OK;
    }

    if (remoteOn != 1) {
        logger(L_INF, "Remote is off. Skip");
        return EXIT_OK;
    }

    switch (ci->type) {	// Non parametrized commands

    case ID_EXIT:
        logger(L_INF, "Command: Exit");
        sendAbort();
        return EXIT_OK;

    case ID_TIMER:
        logger(L_INF, "Command: Timer");
        return processTimerCmd(ci);

    default:
        break;
    }

    switch (ci->type) {		// Parametrized commands, will parametrize in handler

    case ID_MAKE:
        //logger(L_INF, "Command: Make");
        return makeCmd(ci, p);

    case ID_EXECSET:
        logger(L_INF, "Command: ExecAndSet");
        if (connectInfo.mode == SERVER_BT  ||
            connectInfo.mode == SERVER_L2CAP ||
            connectInfo.mode == SERVER_TCP ||
            connectInfo.mode == SERVER_WEB ||
            connectInfo.mode == SERVER_UX  ||
            connectInfo.mode == CLIENT_NOAT) {
            execCmdAndGetResults(ID_SET, ci, NULL, p);
            logger(L_DBG, "Command: ExecAndSet FINISHED");
        }
        return EXIT_OK;

    case ID_EXECSEND:
        logger(L_INF, "Command: ExecAndSend");
        if (connectInfo.mode == SERVER_BT  ||
            connectInfo.mode == SERVER_L2CAP ||
            connectInfo.mode == SERVER_TCP ||
            connectInfo.mode == SERVER_UX  ||
            connectInfo.mode == CLIENT_NOAT) {
            execCmdAndGetResults(ID_SEND, ci, NULL, p);
        }
        return EXIT_OK;

    case ID_MACRO:
        logger(L_INF, "Command: Macro");
        return macroCmd(ci->descr,ci->exec,p);

    default:
        break;
    }

    if (ci->type == ID_EXEC) {	// Parametrized command, handler will use already parametrized data
        logger(L_INF, "Command: Exec");

        char* sExec = substParams(ci->exec, p);
        if (sExec == NULL) {
            logger(L_DBG, "Empty command after substParams()");
            return EXIT_NOK;
        }

        executeSimpleCmdNoPipe(sExec);
        free(sExec);

        return EXIT_OK;
    }

    char* sDescr = substParams(ci->descr, p);
    if (sDescr == NULL) {
        logger(L_DBG, "Empty command after substParams()");
        return EXIT_NOK;
    }

    switch (ci->type) {		// Parametrized commands, handler will use already parametrized data

    case ID_SENDCKPD:
        logger(L_INF, "Command: SendCKPD");
        ret = ckpdCmd(sDescr);
        break;

    case ID_SEND:
        logger(L_INF, "Command: Send");
        if (connectInfo.mode == SERVER_BT  ||
            connectInfo.mode == SERVER_L2CAP ||
            connectInfo.mode == SERVER_TCP ||
            connectInfo.mode == SERVER_UX  ||
            connectInfo.mode == CLIENT_NOAT) {
            ret = processSendCmd(sDescr);
        }
        break;

    case ID_SET:
        logger(L_INF, "Command: Set");
        if (sDescr != NULL &&
                (connectInfo.mode == SERVER_BT   ||
                 connectInfo.mode == SERVER_L2CAP ||
                 connectInfo.mode == SERVER_TCP  ||
                 connectInfo.mode == SERVER_WEB  ||
                 connectInfo.mode == SERVER_UX   ||
                 connectInfo.mode == CLIENT_NOAT
                )) {

            strcpy(tmptext, CMD_SET);
            strcat(tmptext, "(");

            //sprintf(tmp, "ID_SET >%s<", sDescr);
            //logger(L_DBG, tmp);

            if (strstr(sDescr,UPLOAD_ICON)  == sDescr || 	// handle Set(cover|image,icon|window,[...]_image_file_name_); separately.
                strstr(sDescr,UPLOAD_PIX)   == sDescr ||
                strstr(sDescr,UPLOAD_COVER) == sDescr) {

                uploadPix(sDescr);

            } else {

                if (isDataNew(cmdSet2id(sDescr), sDescr, strlen(sDescr))) {
                    logger(L_INF, "Skip to send the same data");
                } else {
                    logger(L_INF, "Data is new");

                    // Change \n in input to real '\n'
                    char *p1 = sDescr;
                    char *p2 = sDescr;
                    while (p2 != NULL) {
                        p2 = strstr(p1, "\\n");

                        if (p2 == NULL) {
                            strncat(tmptext, p1, p2-p1);
                        } else {
                            strncat(tmptext, p1, p2-p1);
                            strcat(tmptext, "\n");
                        }
                        p1 = p2 + 2;
                    }
                    strcat(tmptext, ");");

                    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
                    dm->type  = DM_SET;
                    dm->size  = strlen(tmptext);
                    dm->value = (void*) strdup(tmptext);
		    dm->file  = NULL;

                    sendToDispatcher(dm);
                }
            }
        }
        break;

    case ID_GET:
        logger(L_INF, "Command: Get");
        ret = getCmd(sDescr);
        break;

    case ID_LOAD:
        logger(L_INF, "Command: Load");
        ret = loadCmds(sDescr);
        break;

    case ID_INCLUDE:
        logger(L_INF, "Command: Include");
        ret = includeCmd(sDescr);
        break;

    case ID_EMU:
        logger(L_INF, "Command: Emulate");
        ret = emulateCommands(sDescr);
        break;

    case ID_DBUS:
        logger(L_INF, "Command: Dbus");
        ret = dbusCommand(sDescr);
        break;

    default:
        logger(L_INF, "Unknown command");
        break;
    }
    free(sDescr);
    return ret;
}

int processCommands(cmdItem* cmdList,cmdParams* p)
{
    while (cmdList) {
        /*if (cmdList->descr == NULL) {
                    sprintf(tmp, "Process next command NULL parameters");
            } else {
                    sprintf(tmp, "Process next command %s",cmdList->descr);
            }
            logger(L_DBG, tmp);
            */
        int ret = processOneCommand(cmdList, p);
        if (ret != EXIT_OK) {
            //sprintf(tmp, "Return code from processOneCommand() is %d",ret);
            //logger(L_DBG, tmp);

            if (ret == EXIT_STOP) {
                logger(L_DBG, "Stop execution on current command sequence");
                return EXIT_STOP; 	// just stop execution of command sequence
            }
        }

        cmdList = cmdList->next;
    }

    return EXIT_OK;
}

//
// Top-level command processing function
//
int handleCmdByKey(type_key* k, cmdParams* p)
{
    return handleCmdByKeyEx(k, p, 1);
}

int handleCmdByKeyEx(type_key* k, cmdParams* p, int sendToMain)
{
    if (k == NULL) {
        logger(L_DBG, "handleCmdByKey() NULL input");
        return EXIT_NOK;
    }
    int ret =  EXIT_OK;

    sprintf(tmp, "handleCmdByKey() >%s<",k->key);
    logger(L_DBG, tmp);

    cmdItem* cmdList = getCommand(k);
    if (cmdList == NULL) {
        logger(L_DBG, "No appropriate command was found by getCommand()");
    } else {
        //sprintf(tmp, "ready to execute command");
        //logger(L_DBG, tmp);

        ret = processCommands(cmdList, p);

        // In AT mode: User had pressed some key ...
        // Send ToMainMenu sequence of CKPD's to show main menu again
        // Do we send it even if remote is off ?
        if (remoteOn == 1 &&
                (connectInfo.mode == CLIENT_RFCOMM || connectInfo.mode == CLIENT_AT) &&
                sendToMain == 1) {
            sendToMainMenu();
        }
    }

    return ret;
}

